unit main;
{$WARN SYMBOL_PLATFORM OFF}
{$WARN UNIT_PLATFORM OFF}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, ExtCtrls, ShellCtrls, StdCtrls, ShlObj,
  rkShellTreeview, rkShellListView, rkPathViewer, rkShellPath;

type
  TForm1 = class(TForm)
    splMain: TSplitter;
    stbMain: TStatusBar;
    pnlInfo: TPanel;
    btnClose: TButton;
    btnFont: TButton;
    FontDialog1: TFontDialog;
    ShlTreeView: TRKShellTreeView;
    ShlListView: TRkShellListView;
    ShellPath: TrkShellPath;
    PathViewer: TrkPathViewer;
    procedure FormCreate(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
    procedure btnFontClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure ShlTreeViewChange(Sender: TObject; Node: TTreeNode);
    procedure PathViewerPathChanged(Sender: TObject);
    procedure UpdateGUI;
  private
    { Private declarations }
    AIniFileName: string;
    AStartupFolder: string;
    AViewPathChanging: boolean;
    function LoadSettings(AFileName: string): boolean;
    procedure SaveSettings(AFileName: string);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses ActiveX, IniFiles;
{$R *.dfm}

procedure FreePIDL(PIDL: ShlObj.PItemIDList);
{ Uses to shell allocator to free the memory used by a given PIDL. }
var
  Malloc: ActiveX.IMalloc; // shell's allocator
begin
  // Try to get shell allocator
  if Windows.Succeeded(ShlObj.SHGetMalloc(Malloc)) then
    // Use allocator to free PIDL: Malloc is freed by Delphi
    Malloc.Free(PIDL);
end;

function PIDLToFolderPath(PIDL: ShlObj.PItemIDList): string;
// Takes a PIDL and returns a string
begin
  { Set max length of return string }
  SetLength(Result, Windows.MAX_PATH);
  { Get the path }
  if ShlObj.SHGetPathFromIDList(PIDL, PChar(Result)) then
    Result := PChar(Result)
  else
    Result := '';
end;

function PathToPIDL(APath: WideString): PItemIDList;
// Takes the passed Path and attempts to convert it to the equavalent PIDL
var
  Desktop: IShellFolder;
  pchEaten, dwAttributes: ULONG;
begin
  Result := nil;
  SHGetDesktopFolder(Desktop);
  dwAttributes := 0;
  if Assigned(Desktop) then
  begin
    if Desktop.ParseDisplayName(0, nil, PWideChar(APath), pchEaten, Result,
      dwAttributes) <> NOERROR then
      Result := nil
  end
end;

function LocalAppDataFolder: string;
{ Find LocalAppData folder location }
var
  r: Bool;
  Path: array [0 .. MAX_PATH] of Char;
begin
  r := ShGetSpecialFolderPath(0, Path, CSIDL_LOCAL_APPDATA, False);
  if not r then
    raise Exception.Create('Could not find LocalAppData folder location.');
  Result := Path;
end;

function AppDataFolder: string;
{ Find AppData folder location }
var
  r: Bool;
  Path: array [0 .. MAX_PATH] of Char;
begin
  r := ShGetSpecialFolderPath(0, Path, CSIDL_APPDATA, False);
  if not r then
    raise Exception.Create('Could not find AppData folder location.');
  Result := Path;
end;

function TForm1.LoadSettings(AFileName: string): boolean;
var
  Ini: TIniFile;
begin
  Ini := TIniFile.Create(AFileName);
  try
    Left := Ini.ReadInteger('Form', 'Left', 0);
    Top := Ini.ReadInteger('Form', 'Top', 0);
    Width := Ini.ReadInteger('Form', 'Width', 850);
    Height := Ini.ReadInteger('Form', 'Height', 400);
    WindowState := TWindowState(Ini.ReadInteger('Form', 'Window State', 0));
    FontDialog1.Font.Name := Ini.ReadString('Form', 'FontName', 'Segoe UI');
    FontDialog1.Font.Size := Ini.ReadInteger('Form', 'FontSize', 12);
    FontDialog1.Font.Color := Ini.ReadInteger('Form', 'FontColor', clBlack);
    Form1.Font := FontDialog1.Font;
    AStartupFolder := Ini.ReadString('Form', 'StartupFolder', '');
    Result := True;
  finally
    Ini.Free;
  end;
end;

procedure TForm1.PathViewerPathChanged(Sender: TObject);
begin
  AViewPathChanging := True;
  ShlTreeView.SetPathFromPIDL(ShellPath.PIDL);
  AViewPathChanging := False;
  UpdateGUI;
end;

procedure TForm1.ShlTreeViewChange(Sender: TObject; Node: TTreeNode);
begin
  if not AViewPathChanging then
    ShellPath.LoadPIDL(ShlTreeView.SelectedFolder.AbsoluteID);
  UpdateGUI;
end;

procedure TForm1.UpdateGUI;
begin
  Caption:= ShellPath.Path + ' - Shell PIDL sample';
end;

procedure TForm1.SaveSettings(AFileName: string);
var
  Ini: TIniFile;
begin
  Ini := TIniFile.Create(AFileName);
  try
    Ini.WriteInteger('Form', 'Left', Left);
    Ini.WriteInteger('Form', 'Top', Top);
    Ini.WriteInteger('Form', 'Width', Width);
    Ini.WriteInteger('Form', 'Height', Height);
    Ini.WriteInteger('Form', 'Window State', Ord(WindowState));
    Ini.WriteString('Form', 'FontName', FontDialog1.Font.Name);
    Ini.WriteInteger('Form', 'FontSize', FontDialog1.Font.Size);
    Ini.WriteInteger('Form', 'FontColor', FontDialog1.Font.Color);
    AStartupFolder := ShlTreeView.Path;
    Ini.WriteString('Form', 'StartupFolder', AStartupFolder);
  finally
    Ini.Free;
  end;
end;

procedure TForm1.btnCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TForm1.btnFontClick(Sender: TObject);
begin
  FontDialog1.Font := Form1.Font;
  if FontDialog1.Execute then
  begin
    Form1.Font := FontDialog1.Font;
    PathViewer.Font := FontDialog1.Font;
    ShlTreeView.Font := FontDialog1.Font;
    ShlListView.Font := FontDialog1.Font;
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  AViewPathChanging := False;
  AIniFileName := AppDataFolder + '\ASG\RKViewPathDemo\RKViewPathDemo.ini';
  ForceDirectories(AppDataFolder + '\ASG\RKViewPathDemo');
  if FileExists(AIniFileName) then
    LoadSettings(AIniFileName);
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  SaveSettings(AIniFileName);
end;

procedure TForm1.FormShow(Sender: TObject);
begin
  if DirectoryExists(AStartupFolder) then
    ShlTreeView.Path := AStartupFolder;
end;

end.
